#include <common.h>
#include <exports.h>
#include <net.h>
//#include <nand.h>
//#include <jffs2/jffs2.h>
#include <stdio_dev.h>
#include <asm/errno.h>


// discovery protocol tag values
#define TagNoop              0
#define TagDiscRequest       1
#define TagDiscReply         2
#define TagAnnounce          3
#define TagRebootRequest     4
#define TagRebootReply       5
#define TagNetRestartRequest 6
#define TagNetRestartReply   7
#define TagSetIpRequest      8
#define TagSetIpReply        9
#define TagLedFlashRequest  10
#define TagLedFlashReply    11

#define TagSerial           16
#define TagAck              17
#define TagManufacturer     18
#define TagModel            19
#define TagVersion          20
#define TagUID              21
#define TagMACAddr          22
#define TagStaticIPv4Addr   23
#define TagStaticIPv4Mask   24
#define TagStaticIPv4Gate   25
#define TagDHCPEnable       26
#define TagIPv4LLEnable     27
#define TagCurrentIPv4Addr  28
#define TagCurrentIPv4Mask  29
#define TagCurrentIPv4Gate  30
#define TagStatus           31
#define TagNumEthPorts      32
#define TagNumSerialPorts   33
#define TagSecureConfig     34
#define TagHostname         35
#define TagMessage          36
#define TagLedFlashMode     37
#define TagSwitchPosition   38

static int dhcpEnabled(void)
{
  char *s = getenv("ipmode");
  return s && !strcmp(s,"dhcp");
}

static char *mySerial;

// If there's a serial option and it's not my serial, return false.
// Otherwise, return true.
static int packetForMe(uint8_t *buf, unsigned len)
{
  uint8_t *p = buf;
  while (p < buf+len-1)
    {
      if (p[0] == TagSerial)
        return (p[1] == strlen(mySerial)) && !memcmp(mySerial,&p[2],p[1]);
     p += p[1]+2;
    }
  return 1;
}

// Functions used to add tag/len/option tuples to a packet pointed to
// by 'p'.  Return number of bytes stored.

static int addOptionBytes(uint8_t *p, uint8_t tag, const uint8_t *data, unsigned len)
{
  p[0] = tag;
  p[1] = len;
  memcpy(p+2,data,len);
  return len+2;
}

static int addOptionUnsigned8(uint8_t *p, uint8_t tag, const uint8_t data)
{
  return addOptionBytes(p,tag, &data, sizeof data);
}

#if 0
static int addOptionUnsigned16(uint8_t *p, uint8_t tag, const uint16_t data)
{
  return addOptionBytes(p,tag,(const uint8_t *)&data, sizeof data);
}
#endif

static int addOptionUnsigned32(uint8_t *p, uint8_t tag, const uint32_t data)
{
  return addOptionBytes(p,tag,(const uint8_t *)&data, sizeof data);
}

static int addOptionString(uint8_t *p, uint8_t tag, const char *data)
{
  p[0] = tag;
  if (data == NULL)
    p[1] = 0;
  else
    {
      p[1] = strlen(data);
      memcpy(p+2,data,p[1]);
    }
  return p[1]+2;
}

static int doReboot;

static int handleRebootRequest(uchar *rsp, uchar *req, unsigned len)
{
  printf("reboot request\n");
  uchar *p = rsp;

  p += addOptionBytes(p,TagRebootReply,NULL,0);
  p += addOptionString(p,TagSerial,mySerial);
  p += addOptionUnsigned8(p,TagAck,1);

  doReboot = 1;

  return p-rsp;
}

static int handleNetRestartRequest(uchar *rsp, uchar *req, unsigned len)
{
  printf("net restart request\n");
  uchar *p = rsp;

  p += addOptionBytes(p,TagNetRestartReply,NULL,0);
  p += addOptionString(p,TagSerial,mySerial);
  p += addOptionUnsigned8(p,TagAck,1);
  p += addOptionString(p,TagMessage,"Net restart not supported -- will reboot.");
  doReboot = 1;

  return p-rsp;
}

static int handleLedFlashRequest(uchar *rsp, uchar *req, unsigned len)
{
  printf("LED flash request\n");
  uchar *p = rsp;
  uchar *r;
  char tmpbuf[128];
  uint8_t  flashMode = 0xff;

  p += addOptionBytes(p,TagLedFlashReply,NULL,0);
  p += addOptionString(p,TagSerial,mySerial);

  // verify request
  
  for (r=req+2+req[1];  r < req+len; r += 2+r[1])
    {
      uint8_t  tag = r[0];
      uint8_t  len = r[1];
      uint8_t *data = &r[2];

      if (tag == TagLedFlashMode)
        {
          if (len == 1)
            flashMode = data[0] != 0;
          else
            {
              sprintf(tmpbuf, "ERR: opt %d bad len\n", tag);
              p += addOptionString(p, TagMessage, tmpbuf);
            }
        }
    }
      
  if (flashMode != 0xff)
    {
      extern unsigned comtrol_LED_flash_mode;
      printf("flashMode: %d\n",flashMode);
      comtrol_LED_flash_mode = flashMode;
    }
  else
    {
      sprintf(tmpbuf, "ERR: no valid flash mode option found\n");
      p += addOptionString(p, TagMessage, tmpbuf);
      p += addOptionUnsigned8(p, TagAck, 0);
    }

  return p-rsp;
}


static int handleSetIpRequest(uchar *rsp, uchar *req, unsigned len)
{
  printf("set ip request\n");
  uchar *p = rsp;
  uchar *r;
  char tmpbuf[128];
  int nak = 0;

  p += addOptionBytes(p,TagSetIpReply,NULL,0);
  p += addOptionString(p,TagSerial,mySerial);

  // verify request
  
  for (r=req+2+req[1];  r < req+len; r += 2+r[1])
    {
      uint8_t  tag = r[0];
      uint8_t  len = r[1];
      uint8_t *data = &r[2];

      switch (tag)
        {
        case TagStaticIPv4Addr:
          if (len != 4)
            goto badlen;
          break;

        case TagStaticIPv4Mask:
          if (len != 4)
            goto badlen;
          break;

        case TagStaticIPv4Gate:
          if (len != 4)
            goto badlen;
          break;

        case TagDHCPEnable:
          if (len != 1)
            goto badlen;
          break;

        case TagIPv4LLEnable:
          if (len != 1)
            goto badlen;
          if (data[0] != 0)
            goto badval;
          break;

        case TagHostname:
          break;

        case TagSerial:
        case TagNoop:
          continue;

        default:
          sprintf(tmpbuf, "WARN: opt %d not impl", tag);
          p += addOptionString(p, TagMessage, tmpbuf);
          continue;
        }

      continue;

    badlen:
      sprintf(tmpbuf, "ERR: opt %d bad len", tag);
      p += addOptionString(p, TagMessage, tmpbuf);
      nak = 1;
      continue;

    badval:
      sprintf(tmpbuf, "ERR: opt %d bad val", tag);
      p += addOptionString(p, TagMessage, tmpbuf);
      nak = 1;
      continue;
    }

  if (!nak)
    {
      // process request

      for (r=req+2+req[1];  r < req+len; r += 2+r[1])
        {
          uint8_t  tag = r[0];
          uint8_t *data = &r[2];
          char tmpbuf[128];
          unsigned a;
      
          switch (tag)
            {
            case TagStaticIPv4Addr:
              memcpy(&a, data, 4);
              ip_to_string(a,tmpbuf);
              setenv("ipaddr",tmpbuf);
              break;
            case TagStaticIPv4Mask:
              memcpy(&a, data, 4);
              ip_to_string(a,tmpbuf);
              setenv("netmask",tmpbuf);
              break;
            case TagStaticIPv4Gate:
              memcpy(&a, data, 4);
              ip_to_string(a,tmpbuf);
              setenv("gatewayip",tmpbuf);
              break;
            case TagDHCPEnable:
              setenv("ipmode", data[0] ? "dhcp" : "static");
              break;
            case TagHostname:
              {
                uint8_t  len = r[1];
                if (len >= sizeof tmpbuf)
                  len = (sizeof tmpbuf)-1;
                memcpy(tmpbuf,data,len);
                tmpbuf[len] = '\0';
                setenv("hostname",tmpbuf);
              }
              break;
            default:
              break;
            }

          saveenv();
        }

      p += addOptionString(p,TagHostname,getenv("hostname"));
      
      p += addOptionUnsigned32(p,TagStaticIPv4Addr,getenv_IPaddr("ipaddr"));
      p += addOptionUnsigned32(p,TagStaticIPv4Mask,getenv_IPaddr("netmask"));
      p += addOptionUnsigned32(p,TagStaticIPv4Gate,getenv_IPaddr("gatewayip"));
      
      p += addOptionUnsigned32(p,TagCurrentIPv4Addr,NetOurIP);
      p += addOptionUnsigned32(p,TagCurrentIPv4Mask,NetOurSubnetMask);
      p += addOptionUnsigned32(p,TagCurrentIPv4Gate,NetOurGatewayIP);
      
      p += addOptionUnsigned8(p,TagSecureConfig,0);
      p += addOptionUnsigned8(p,TagDHCPEnable,dhcpEnabled());
      p += addOptionUnsigned8(p,TagIPv4LLEnable,0);
    }


  p += addOptionUnsigned8(p, TagAck, !nak);

  return p-rsp;
}

static char *switchPosString(void)
{
  static char s[4];
  uint16_t p = *((volatile uint16_t *)(0x10000c40));
  p = (p&0x0f00)>>8 | (p&0x00f0) | (p&0x000f)<<8;
  sprintf(s,"%03x",p);
  return s;
}

extern const char version_string[];

static int handleDiscRequest(uchar *rsp, uchar *req, unsigned len)
{
  printf("discovery request\n");

  uchar *p = rsp;

  p += addOptionBytes(p,TagDiscReply,NULL,0);
  p += addOptionString(p,TagSerial,mySerial);
  p += addOptionBytes(p, TagMACAddr, NetOurEther, 6);

  p += addOptionString(p,TagManufacturer,getenv("manufacturer"));
  p += addOptionString(p,TagModel,getenv("model"));
  p += addOptionString(p,TagVersion,version_string);

  p += addOptionString(p,TagHostname,getenv("hostname"));

  p += addOptionUnsigned32(p,TagStaticIPv4Addr,getenv_IPaddr("ipaddr"));
  p += addOptionUnsigned32(p,TagStaticIPv4Mask,getenv_IPaddr("netmask"));
  p += addOptionUnsigned32(p,TagStaticIPv4Gate,getenv_IPaddr("gateway"));

  p += addOptionUnsigned32(p,TagCurrentIPv4Addr,NetOurIP);
  p += addOptionUnsigned32(p,TagCurrentIPv4Mask,NetOurSubnetMask);
  p += addOptionUnsigned32(p,TagCurrentIPv4Gate,NetOurGatewayIP);

  p += addOptionUnsigned8(p,TagSecureConfig,0);
  p += addOptionUnsigned8(p,TagDHCPEnable,dhcpEnabled());
  p += addOptionUnsigned8(p,TagIPv4LLEnable,0);

  p += addOptionString(p,TagSwitchPosition,switchPosString());

  return p-rsp;
}


// Number of bytes in header (cookie + transaction ID).  First tag
// starts at this offset.
#define HdrSize 12


#define isBroadcastEther(a) ((a)[0] == 0xff && (a)[1] == 0xff && (a)[2] == 0xff && (a)[3] == 0xff && (a)[4] == 0xff && (a)[5] == 0xff)

void handleDiscoveryPacket(uchar *pkt, unsigned dstPort, unsigned srcPort, unsigned len)
{
  uint8_t  srcEther[6];
  uint8_t  dstEther[6];
  uint32_t srcIP;
  uint32_t dstIP;
  IP_t *ip;
  int bytecount;
  uchar *rsp = (uchar*)NetTxPacket + NetEthHdrSize() + IP_HDR_SIZE;

  ip = (IP_t*)(pkt-IP_HDR_SIZE);
  memcpy(&srcIP,&ip->ip_src,sizeof srcIP);
  memcpy(&dstIP,&ip->ip_dst,sizeof dstIP);
  memcpy(srcEther,(pkt-IP_HDR_SIZE-NetEthHdrSize())+6,6);
  memcpy(dstEther,(pkt-IP_HDR_SIZE-NetEthHdrSize())+0,6);

  printf("got discovery packet\n");
  printf(" from %u.%u.%u.%u:%d  ",(srcIP>>0)&0xff,(srcIP>>8)&0xff,(srcIP>>16)&0xff,(srcIP>>24)&0xff,srcPort);
  printf("[%02x:%02x:%02x:%02x:%02x:%02x]\n",srcEther[0],srcEther[1],srcEther[2],srcEther[3],srcEther[4],srcEther[5]);
  printf("   to %u.%u.%u.%u:%d  ",(dstIP>>0)&0xff,(dstIP>>8)&0xff,(dstIP>>16)&0xff,(dstIP>>24)&0xff,dstPort);
  printf("[%02x:%02x:%02x:%02x:%02x:%02x]\n",dstEther[0],dstEther[1],dstEther[2],dstEther[3],dstEther[4],dstEther[5]);
  printf("  len %d\n", len);

  if (len < HdrSize)
    return;

  mySerial = getenv("serial#");

  memcpy(rsp, pkt, HdrSize);

  // skip over headers for rest of handling
  len -= HdrSize;
  pkt += HdrSize;
  rsp += HdrSize;

  // check for serial number tag and match
  if (!packetForMe(pkt, len))
    return;

  // first tag is required to be request tag -- always immediately
  // following header bytes.  dispatch request to appropriate handler
  
  switch (pkt[0])
    {
    case TagDiscRequest:
      bytecount = handleDiscRequest(rsp, pkt, len);
      break;

    case TagNetRestartRequest:
      bytecount = handleNetRestartRequest(rsp, pkt, len);
      break;

    case TagRebootRequest:
      bytecount = handleRebootRequest(rsp, pkt, len);
      break;

    case TagSetIpRequest:
      bytecount = handleSetIpRequest(rsp, pkt, len);
      break;

    case TagLedFlashRequest:
      bytecount = handleLedFlashRequest(rsp, pkt, len);
      break;

    default:
      // ignore it
      bytecount = 0;
      break;
    }

  if (bytecount)
    {
      bytecount += HdrSize;

      if (isBroadcastEther(dstEther))
        NetSendUDPPacket(dstEther, 0xffffffff, srcPort, 4606, bytecount);
      else
        NetSendUDPPacket(srcEther, srcIP, srcPort, 4606, bytecount);
    }

  if (doReboot)
    {
      volatile unsigned u;
      for (u=0; u<100000; ++u)
        ;
      run_command("reset",0);
    }
}


